/* 
 * FILE --- fuzzy_sys.cc
 *
 * PURPOSE --- Source file for a class for a multiple input multiple output 
 * (MIMO) standard fuzzy system. The user specifies the # of inputs and
 * outputs as well as the # of membership functions (MF) on each universe 
 * of discourse (UD). The MF can be defined as triangular with 50% of 
 * overlapp or as gaussian. Either minimum or product can be used for 
 * premises, and center of gravity (COG) or center average (CAV) for 
 * defuzzification.
 *
 * Uses two other utilities:
 *                  FUZZY_UD --- Fuzzy universe of discourse
 *                  FUZZY_MF --- Fuzzy membership function
 *
 */

// Include Files
#include <math.h>
#include "fuzzy_sys.hh"

// Define this macro if you want to use this utility without linking
// to the RCS Library
#ifdef NO_RCS
#include <stdio.h>            /* printf function */
#else
#include "rcs_prnt.hh"          /* rcs_print_ functions */
#endif


/*************************************************************/

FUZZY_SYS::FUZZY_SYS(int _num_of_inputs, 
		     int _num_of_outputs,
		     int *_num_of_in_mfs = NULL, 
		     int *_num_of_out_mfs = NULL)
{
  // No error at start
  fuzzy_sys_err = FUZZY_SYS_NO_ERROR;
  num_of_inputs = _num_of_inputs;
  num_of_outputs = _num_of_outputs;

  // Allocate memory 
  allocate_memory(_num_of_in_mfs, _num_of_out_mfs);

  // If error then report
  if(fuzzy_sys_err != FUZZY_SYS_NO_ERROR)
    {
#ifdef NO_RCS
      printf("FUZZY_SYS: ERROR --- Not Enought Memory!!!\n");
      printf("The Fuzzy System will not function correctly ...\n");
#else
      rcs_print("FUZZY_SYS: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The Fuzzy System will not function correctly ...\n");
#endif
    }

  // Calculate the # of rules 
  num_of_rules = 1;
  for (int ii=0; ii < num_of_inputs; ii++)
    {
      num_of_rules *= in_ud[ii]->get_num_of_mfs();
    }

  // Set the the default premise type
  set_premise_type();

  // Set the the default defuzzification type
  set_defuzz_type();

  // Set the input gains to zeros
  set_input_gains();

  // Set the output gains to zeros
  set_output_gains();
}


// Destructor
FUZZY_SYS::~FUZZY_SYS()
{
  // Free the allocated memory
  free_memory();
}

void FUZZY_SYS::free_memory(void)
{
  int ii;  // Index variales for the loops

  for (ii=0; ii < num_of_inputs; ii++)
    {
     delete in_ud[ii];
    }
  delete [] in_ud;

  for (ii=0; ii < num_of_outputs; ii++)
    {
     delete out_ud[ii];
    }
  delete [] out_ud;

  delete [] total_area;
  delete [] weighted_total_area;
}


void FUZZY_SYS::allocate_memory(int *_num_of_in_mfs, int *_num_of_out_mfs)
{
  int ii;  // Index variales for the loops

  // Input MF 
  if ((in_ud = new fuzzy_ud*[num_of_inputs]) == NULL)
    fuzzy_sys_err = FUZZY_SYS_MEM_ERROR;
  if (_num_of_in_mfs == NULL) // No num of MF is specified
    {
      for (ii=0; ii < num_of_inputs; ii++)
	{
	  in_ud[ii] = new fuzzy_ud(DEFAULT_NUM_IN_MF);
	  if (in_ud[ii]->fuzzy_ud_err != FUZZY_UD_NO_ERROR)
	    fuzzy_sys_err = FUZZY_SYS_MEM_ERROR;
	}
    }
  else  // Num of MF is specified
    {
      for (ii=0; ii < num_of_inputs; ii++)
	{
	  in_ud[ii] =  new fuzzy_ud(_num_of_in_mfs[ii]);
	  if (in_ud[ii]->fuzzy_ud_err != FUZZY_UD_NO_ERROR)
	    fuzzy_sys_err = FUZZY_SYS_MEM_ERROR;
	}
    }

  // Output MF 
  if ((out_ud = new fuzzy_ud*[num_of_outputs]) == NULL)
    fuzzy_sys_err = FUZZY_SYS_MEM_ERROR;
  if (_num_of_out_mfs == NULL) // No num of MF is specified
    {
      for (ii=0; ii < num_of_outputs; ii++)
	{
	  out_ud[ii] = new fuzzy_ud(DEFAULT_NUM_OUT_MF);
	  if (out_ud[ii]->fuzzy_ud_err != FUZZY_UD_NO_ERROR)
	    fuzzy_sys_err = FUZZY_SYS_MEM_ERROR;
	}
    }
  else  // Num of MF is specified
    {
      for (ii=0; ii < num_of_outputs; ii++)
	{
	  out_ud[ii] =  new fuzzy_ud(_num_of_out_mfs[ii]);
	  if (out_ud[ii]->fuzzy_ud_err != FUZZY_UD_NO_ERROR)
	    fuzzy_sys_err = FUZZY_SYS_MEM_ERROR;
	}
    }

  // Conveniance variable
  if ((total_area = new double[num_of_outputs]) == NULL)
    fuzzy_sys_err = FUZZY_SYS_MEM_ERROR;

  // Conveniance variable
  if ((weighted_total_area = new double[num_of_outputs]) == NULL)
    fuzzy_sys_err = FUZZY_SYS_MEM_ERROR;
}

// This function sets the premise type. Invalid premices are considered
// as MINIMUM_PREMISE.
void FUZZY_SYS::set_premise_type(int _premise_type = MINIMUM_PREMISE)
{
  premise_type = _premise_type;
  if((premise_type != MINIMUM_PREMISE) && (premise_type != PRODUCT_PREMISE))
    {
#ifdef NO_RCS
      printf("Invalid premise type!!!\n");
      printf("Minimum premise assumed ...\n");
#else
      rcs_print("Invalid premise type!!!\n");
      rcs_print("Minimum premise assumed ...\n");
#endif
      premise_type = MINIMUM_PREMISE;      
    }
}

// This function sets the defuzzyfication type. Invalid defuzzyfication
// types are considered as COG_DEFUZZ.
void FUZZY_SYS::set_defuzz_type(int _defuzz_type = COG_DEFUZZ)
{
  defuzz_type = _defuzz_type;
  if((defuzz_type != COG_DEFUZZ) && (defuzz_type != CAV_DEFUZZ))
    {
#ifdef NO_RCS
      printf("Invalid defuzzyfication type!!!\n");
      printf("Center-of-Gravity defuzzyfication assumed ...\n");
#else
      rcs_print("Invalid defuzzyfication type!!!\n");
      rcs_print("Center-of-Gravity defuzzyfication assumed ...\n");
#endif
      defuzz_type = COG_DEFUZZ;
    }
}

void FUZZY_SYS::set_input_gains(double *_in_gains = NULL)
{
  int ii;  // Index variales for the loops

  // Set the gains for each input universe
  if (_in_gains == NULL) // No gains specified
    {
      for (ii=0; ii < num_of_inputs; ii++)
	{
	  in_ud[ii]->ud_gain = 1.0;
	}
    }
  else // Gains specified
    {
      for (ii=0; ii < num_of_inputs; ii++)
	{
	  in_ud[ii]->ud_gain = _in_gains[ii];
	}
    }
}

void FUZZY_SYS::set_output_gains(double *_out_gains = NULL)
{
  int ii;  // Index variales for the loops

  // Set the gains for each output universe
  if (_out_gains == NULL) // No gains specified
    {
      for (ii=0; ii < num_of_outputs; ii++)
	{
	  out_ud[ii]->ud_gain = 1.0;
	}
    }
  else // Gains specified
    {
      for (ii=0; ii < num_of_outputs; ii++)
	{
	  out_ud[ii]->ud_gain = _out_gains[ii];
	}
    }
}

/*
  The function that given an array of crisp inputs evaluates the crisp 
  output of the fuzzy controller. This is the only function to be called 
  from outside.
  input = an array of crisp inputs to the fuzzy system
  output = the crisp outputs of the fuzzy system
*/
void FUZZY_SYS::control(const double *input, double *output)
{
  // Scale and saturate the input to the correct universe
  for (int ii=0; ii < num_of_inputs; ii++)
    {
      in_ud[ii]->fuzzyfy(input[ii]);
    }

  // Evaluate and return the crisp outputs
  defuzzyfy(output);
}  

void FUZZY_SYS::defuzzyfy(double *output)
{
  int ii, jj, kk;  // Index variales for the loops
  int input_flag, on_off_index, mf_index, out_mf_index;
  double inference;
  double area;

  // Start with zero sum 
  for (ii=0; ii < num_of_outputs; ii++)
    {
      total_area[ii] = 0.0;
      weighted_total_area[ii] = 0.0;
    }	  
  
  /*
    The algorithm below evaluates all the possible rules.
    With the help of input_flag and on_off_index we find
    which fuzzy set (i.e., MF) from which input belongs
    to each rule.
   */
  for (jj=0; jj < num_of_rules; jj++)
    {
      input_flag = jj; // In each rule belongs different MF from each UD
      inference = 1.0; // Maximum possible inference
      mf_index = 0;  // Output MF is unknown
      
      for (kk=0; kk < num_of_inputs; kk++) 
	{
	  on_off_index = (input_flag % in_ud[kk]->get_num_of_mfs());
	  input_flag = 
	    (input_flag - on_off_index)/in_ud[kk]->get_num_of_mfs();
	  
	  // Find the inference
	  switch(premise_type)
	    {
	    case MINIMUM_PREMISE:
	      inference = 
		MIN(in_ud[kk]->deg_of_mbrship[on_off_index], inference);
	      break;
	    case PRODUCT_PREMISE:
	      inference *= in_ud[kk]->deg_of_mbrship[on_off_index];
	      break;
	    default:
	      // This wont happen, but just in case ...
	      // Play Safe --- Consider every other type as MINIMUM_PREMISE
	      inference = 
		MIN(in_ud[kk]->deg_of_mbrship[on_off_index], inference);
	    }

	  // Use index adding to find the # of the output MF.
	  mf_index += 
	    on_off_index - (in_ud[kk]->get_num_of_mfs() - 1)/2;
	  
	}
      
      // For each output find the output MF and calculate the contribution 
      // of the current rule to this output.
      for (ii=0; ii < num_of_outputs; ii++)
	{
	  // Adjust the mf_index if out of range
	  out_mf_index = SATURATE(mf_index, 
				  (out_ud[ii]->get_num_of_mfs()-1)/2);
	  out_mf_index += (out_ud[ii]->get_num_of_mfs()-1)/2;
	  
	  // Calculate the contribution of this rule to the output
	  switch(defuzz_type)
	    {
	    case COG_DEFUZZ:
	      area = 2*out_ud[ii]->mfs[out_mf_index].get_spread()*
		(inference - (inference*inference)/2);
	      break;
	      
	    case CAV_DEFUZZ:
	      area = inference;
	      break;
	      
	    default:
	      // This wont happen, but just in case ...
	      // Play Safe --- Consider every other type as COG_DEFUZZ
	      area = 2*out_ud[ii]->mfs[out_mf_index].get_spread()*
		(inference - (inference*inference)/2);
	    }
	  
	  // Add up the contribution of all the rules
	  total_area[ii] += area;
	  weighted_total_area[ii] += 
	    area*out_ud[ii]->mfs[out_mf_index].get_center();
	}
    }

  // Find each crisp output
  for (ii=0; ii < num_of_outputs; ii++)
    {
      if (total_area[ii] <= 0)
	{
#ifdef NO_RCS
      printf("Inproper fuzzy system!!!\n");
      printf("Sum of premise certainties equal to zero!!!\n");
      printf("Setting output %d to zero ...\n", ii);
#else
      rcs_print("Inproper fuzzy system!!!\n");
      rcs_print("Sum of premise certainties equal to zero!!!\n");
      rcs_print("Setting output %d to zero ...\n", ii);
#endif
      output[ii] = 0.0;
	}
      else
	{
	  output[ii] = 
	    MODE*out_ud[ii]->ud_gain*(weighted_total_area[ii]/total_area[ii]);
	}
    }  
}
  
  



